@@ -1,4 +1,5 @@
-import io
+# import io
+from io import BytesIO
 from dataclasses import dataclass, field
 from numbers import Number
 from pathlib import Path
@@ -9,16 +10,27 @@ import numpy as np
 import pandas as pd
 import panel as pn
 import plotly.express as px
+import plotly.io as pio
+from bokeh.colors import RGB
 from bokeh.core.enums import MarkerType
 from bokeh.models import HoverTool
 from bokeh.settings import settings
 from holoviews import opts, dim
 from holoviews.streams import PlotReset, SelectionExpr
+from matplotlib import cm
 from shapely.geometry import Point, MultiPoint, Polygon, MultiPolygon
 from sklearn.decomposition import PCA
 
 from pyhard.context import Workspace
 
+
+pio.templates.default = "plotly"
+m_coolwarm_rgb = (255 * cm.coolwarm(range(256))).astype('int')
+coolwarm_palette = [RGB(*tuple(rgb)).to_hex() for rgb in m_coolwarm_rgb]
+coolwarm_cmap = px.colors.make_colorscale(coolwarm_palette)
+px.defaults.color_continuous_scale = coolwarm_cmap
+px.defaults.color_discrete_sequence = px.colors.qualitative.Plotly
+
 _my_path = Path(__file__).parent
 
 settings.resources = 'cdn'
@@ -27,9 +39,10 @@ settings.resources = 'inline'
 # Palette: https://coolors.co/palette/0466c8-0353a4-023e7d-002855-001845-001233-33415c-5c677d-7d8597-979dac
 dark_color = '#292929'
 light_color = '#F8F8F9'
+transparent = 'rgba(0,0,0,0)'
 footprint_colors = {'good': '#58D68D', 'best': '#9B59B6'}
 
-colorbar_opts = {'background_fill_color': light_color}
+colorbar_opts = {'background_fill_color': transparent}
 grid_opts = {'grid_line_dash': [6, 4], 'grid_line_alpha': .5}
 control_width = 280
 
@@ -65,7 +78,7 @@ class ClassificationApp:
         self.workspace = workspace
         self.bbox = None
         self.cmap = 'coolwarm'
-        self.tabs = pn.Tabs(*[(name, None) for name in self._tabs_labels], dynamic=True)
+        self.tabs = pn.Tabs(*[(name, None) for name in self._tabs_labels], dynamic=False)
 
         self.all_cols = (
                 workspace.data.columns.to_list() +
@@ -91,12 +104,6 @@ class ClassificationApp:
 
         data = df_is.join(df_dataset)
         self.data = data.join(df_metadata)
-        # self.ds = hv.Dataset(self.data.reset_index())
-        # self.selection = link_selections.instance(
-        #     unselected_alpha=0.2,
-        #     # unselected_color='#CBCED5',
-        #     show_regions=False
-        # )
 
         self.data_dims = df_dataset.columns.to_list()
         self.class_label = self.data_dims[2]
@@ -257,11 +264,11 @@ class ClassificationApp:
         Z = self.workspace.is_coordinates.values
         algo = self.w_footprint_algo.value
         try:
-            boundary = self.workspace.footprints.xs([algo, ftype]).values
+            boundary = self.workspace.footprints.xs((algo, ftype)).values
         except KeyError:
             return [False] * Z.shape[0]
         poly = self.footprint2polygons(boundary)
-        return [poly.contains(p) or poly.boundary.contains(p) for p in MultiPoint(Z)]
+        return [poly.contains(p) or poly.boundary.contains(p) for p in MultiPoint(Z).geoms]
 
     def data_space(self, color: str, range_limits: Tuple[Number, Number], autorange_on: bool) -> hv.Scatter:
         """
@@ -425,10 +432,7 @@ class ClassificationApp:
         def download_selection_points():
             mask = self.select_instances()
             df = self.get_download_data(mask)
-            sio = io.StringIO()
-            df.to_csv(sio)
-            sio.seek(0)
-            return sio
+            return BytesIO(df.to_csv().encode())
 
         button_download = pn.widgets.FileDownload(
             embed=False,
@@ -442,10 +446,7 @@ class ClassificationApp:
         def download_footprint_points(ftype: str = 'good'):
             mask = self.select_footprint_instances(ftype)
             df = self.get_download_data(mask)
-            sio = io.StringIO()
-            df.to_csv(sio)
-            sio.seek(0)
-            return sio
+            return BytesIO(df.to_csv().encode())
 
         button_download_fp_good = pn.widgets.FileDownload(
             embed=False,
@@ -466,9 +467,14 @@ class ClassificationApp:
 
         main = (dmap_pca * sel1_dmap + dmap_isa * sel2_dmap).cols(2).opts(
             opts.Layout(shared_axes=False, shared_datasource=True, framewise=True),
-            opts.Scatter(bgcolor='#F8F8F9'),
-            opts.Polygons(show_legend=True, legend_position='bottom_left'),
-            opts.Overlay(legend_opts={"click_policy": "hide", }),
+            opts.Scatter(bgcolor=transparent),
+            opts.Polygons(show_legend=True),
+            opts.Overlay(
+                legend_opts={"click_policy": "hide", },
+                legend_position='bottom',
+                legend_cols=True,
+                legend_offset=(0, 10)
+            ),
         )
 
         control = pn.Column(
@@ -570,7 +576,8 @@ class ClassificationApp:
             )
             fig.layout.autosize = True
             return pn.Column(
-                pn.pane.Plotly(fig),
+                pn.Spacer(),
+                pn.pane.Plotly(fig, aspect_ratio=2),
                 sizing_mode="stretch_width"
             )
 
@@ -604,6 +611,7 @@ class ClassificationApp:
                 x=x,
                 y=y,
                 color=c,
+                symbol=self.output_col,
                 # marginal_x="histogram",
                 # marginal_y="histogram",
                 category_orders=category_orders
@@ -614,10 +622,19 @@ class ClassificationApp:
                 modebar={'bgcolor': 'rgba(255,255,255,0)'},
                 font_family="'Source Sans Pro', verdana, arial",
                 legend_font_size=15,
-                font_size=16
+                font_size=16,
+                legend=dict(
+                    orientation="h",
+                    yanchor="bottom",
+                    y=1.02,
+                    xanchor="left",
+                    x=0
+                )
             )
+            fig_plotly.update_traces(marker=dict(size=7))
             fig_plotly.layout.autosize = True
             return pn.Column(
+                pn.Spacer(),
                 pn.pane.Plotly(fig_plotly, aspect_ratio=2),
                 sizing_mode="stretch_width"
             )
@@ -703,6 +720,229 @@ class ClassificationApp:
                 port += 1
 
 
+class RegressionApp(ClassificationApp):
+    def __int__(self, workspace: Workspace):
+        super().__init__(workspace)
+        delattr(self, 'marker')
+
+    def data_space(self, color: str, range_limits: Tuple[Number, Number], autorange_on: bool) -> hv.Scatter:
+        """
+        Creates the data space (PCA) plot.
+
+        Args:
+            color: name of the variable to color by
+            range_limits: color bar ranges
+            autorange_on: whether to use auto range for color bar
+
+        Returns:
+            Data space scatter plot
+
+        """
+        if not autorange_on:
+            range_limits = (np.nan, np.nan)
+        cmap = self.cmap
+        hover_list = [
+            (self.output_col, self.output_col),
+            ('Component 1', 'Component1'),
+            ('Component 2', 'Component2'),
+            ('Instance Hardness', 'instance_hardness'),
+            (' '.join(str.split(color, '_', 1)), color)
+        ]
+        hover_list = list(dict.fromkeys(hover_list))
+        tooltips = [(s[0], '@' + s[1]) for s in hover_list]
+        hover = HoverTool(tooltips=tooltips)
+        scatter_pca = hv.Scatter(
+            self.data,
+            kdims=self.data_kdims,
+            vdims=self.data_vdims
+        ).opts(
+            responsive=True,
+            aspect=1.1,
+            color=color,
+            cmap=cmap,
+            show_grid=True,
+            gridstyle=grid_opts,
+            tools=['lasso_select', 'box_select', hover],
+            size=7,
+            framewise=True,
+            colorbar=True,
+            colorbar_opts=colorbar_opts,
+            clim=range_limits
+        )
+        return scatter_pca
+
+    def instance_space(self, color: str, range_limits: Tuple[Number, Number], autorange_on: bool) -> hv.Scatter:
+        """
+        Creates the instance space plot.
+
+        Args:
+            color: name of the variable to color by
+            range_limits: color bar ranges
+            autorange_on: whether to use auto range for color bar
+
+        Returns:
+            Instance Space scatter plot
+
+        """
+        if not autorange_on:
+            range_limits = (np.nan, np.nan)
+        cmap = self.cmap
+        hover_list = [
+            (self.output_col, self.output_col),
+            ('z1', 'z_1'),
+            ('z2', 'z_2'),
+            ('Instance Hardness', 'instance_hardness'),
+            (' '.join(str.split(color, '_', 1)), color)
+        ]
+        hover_list = list(dict.fromkeys(hover_list))
+        tooltips = [(s[0], '@' + s[1]) for s in hover_list]
+        hover = HoverTool(tooltips=tooltips)
+        scatter_is = hv.Scatter(
+            self.data,
+            kdims=self.is_kdims,
+            vdims=self.is_vdims
+        ).opts(
+            responsive=True,
+            aspect=1.1,
+            color=color,
+            cmap=cmap,
+            show_grid=True,
+            gridstyle=grid_opts,
+            tools=['lasso_select', 'box_select', hover],
+            size=7,
+            framewise=True,
+            colorbar=True,
+            colorbar_opts=colorbar_opts,
+            clim=range_limits
+        )
+        return scatter_is
+
+    def render_dist_panel(self):
+        @pn.depends(g=self.w_dist_source, watch=True)
+        def update_var_list(g):
+            if g == 'Features':
+                self.w_dist_var.options = self.feat_cols
+            else:
+                self.w_dist_var.options = self.meta_cols
+
+        @pn.depends(
+            a=self.tabs.param.active,
+            dist_type=self.w_dist_type.param.value,
+            var=self.w_dist_var.param.value,
+            source=self.w_dist_source.param.value
+        )
+        def plot_dist(a, dist_type, var, source):
+            if a != 2:
+                return ""
+            mask = self.select_instances()
+            output_col = self.output_col
+
+            if source == 'Features':
+                df_full = self.workspace.data.copy()
+            else:
+                df_full = self.workspace.extended_metadata.copy()
+                df_full = df_full.join(self.workspace.data[output_col])
+
+            df_selection = df_full[mask].copy()
+            # classes = df_selection[output_col].unique().tolist()
+            # classes.sort()
+            # category_orders = {output_col: classes}
+            fig = None
+
+            if dist_type == 'Histogram':
+                fig = px.histogram(
+                    df_selection,
+                    x=var,
+                )
+            elif dist_type == 'Bar':
+                df_count = df_selection[[var]].value_counts()
+                df_count.name = 'count'
+                df_count = df_count.to_frame().reset_index()
+                fig = px.bar(
+                    df_count,
+                    x=var,
+                    y="count",
+                )
+            elif dist_type == 'Box':
+                df_full['Group'] = 'All instances'
+                df_selection['Group'] = 'Selected instances'
+                fig = px.box(
+                    pd.concat([df_selection, df_full]),
+                    x='Group',
+                    y=var,
+                )
+
+            fig.update_layout(
+                margin=dict(l=0, r=0, t=20, b=10),
+                paper_bgcolor='rgba(0,0,0,0)',
+                modebar={'bgcolor': 'rgba(255,255,255,0)'},
+                font_family="'Source Sans Pro', verdana, arial",
+                legend_font_size=14,
+                font_size=16
+            )
+            fig.layout.autosize = True
+            return pn.Column(
+                pn.Spacer(),
+                pn.pane.Plotly(fig, aspect_ratio=2),
+                sizing_mode="stretch_width"
+            )
+
+        self.layout.dists.name = self._tabs_labels[2]
+        self.layout.dists.main = plot_dist
+        self.layout.dists.control = pn.Column(
+            self.w_dist_source,
+            pn.Spacer(),
+            self.w_dist_type,
+            self.w_dist_var,
+            width=control_width
+        )
+
+    def render_explorer_panel(self):
+        @pn.depends(a=self.tabs.param.active, x=self.w_var_x, y=self.w_var_y, c=self.w_var_c.param.value)
+        def plot_scatter(a, x, y, c):
+            if a != 3:
+                return ""
+            mask = self.select_instances()
+            df_all = self.workspace.data[mask].copy()
+            df_all = df_all.join(self.workspace.extended_metadata, how='left')
+            df_all = df_all.join(self.workspace.is_coordinates, how='left')
+
+            fig_plotly = px.scatter(
+                df_all,
+                x=x,
+                y=y,
+                color=c,
+                # marginal_x="histogram",
+                # marginal_y="histogram",
+            )
+            fig_plotly.update_layout(
+                margin=dict(l=0, r=0, t=20, b=10),
+                paper_bgcolor='rgba(0,0,0,0)',
+                modebar={'bgcolor': 'rgba(255,255,255,0)'},
+                font_family="'Source Sans Pro', verdana, arial",
+                legend_font_size=15,
+                font_size=16,
+                legend=dict(
+                    orientation="h",
+                    yanchor="bottom",
+                    y=1.02,
+                    xanchor="left",
+                    x=0
+                )
+            )
+            fig_plotly.update_traces(marker=dict(size=7))
+            fig_plotly.layout.autosize = True
+            return pn.Column(
+                pn.Spacer(),
+                pn.pane.Plotly(fig_plotly, aspect_ratio=2),
+                sizing_mode="stretch_width"
+            )
+
+        self.layout.explorer.name = self._tabs_labels[3]
+        self.layout.explorer.main = plot_scatter
+        self.layout.explorer.control = pn.Column(self.w_var_x, self.w_var_y, self.w_var_c, width=control_width)
+
+
 if __name__ == "__main__":
     ws = Workspace(_my_path / "data/wine/")
     ws.load()
