@@ -6,6 +6,7 @@ import logging
 import os
 import shutil
 import time
+from enum import Enum
 from pathlib import Path
 from typing import Optional
 
@@ -17,12 +18,12 @@ from pyispace.trace import trace_build_wrapper, make_summary, _empty_footprint
 from pyispace.utils import save_footprint, scriptcsv
 
 from . import integrator, formatter
-from .app import ClassificationApp
+from .app import ClassificationApp, RegressionApp
 from .context import Configuration, Workspace
 from .feature_selection import filtfeat, variance_filter, correlation_filter
 from .metrics import loss_threshold
-from .structures import Configurations
-from .utils import pretty_time_delta
+from .structures import Configurations, GeneralConfig, FSConfig, HPOConfig, ISAConfig
+from .utils import pretty_time_delta, check_new_version
 from .validator import *
 from .visualization import Demo
 
@@ -73,12 +74,233 @@ def _validate_paths(config: Configurations) -> Configurations:
 
 
 @cli.callback()
-def configure_log():
+def start_up():
     sh = logging.StreamHandler()
     sh.setLevel(logging.DEBUG)
     sh.setFormatter(formatter)
     logging.getLogger().addHandler(sh)
 
+    if check_new_version("pyhard"):
+        typer.secho(
+            "There is a new version of PyHard available in PyPI. Run the command to update it:\n"
+            "\n\tpip install --upgrade pyhard\n",
+            fg=typer.colors.YELLOW
+        )
+
+
+# @cli.command()
+def wizard():
+    """
+    """
+    typer.echo("Welcome to the PyHard Wizard.\n")
+    typer.echo(
+        "Throughout the next steps the fields of the configuration file will be set."
+    )
+
+    typer.secho("\n# General options #", fg=typer.colors.GREEN)
+    typer.secho(
+        "\nPlease use forward slash ('/') for paths, and also note that they can be relative to the current dir.",
+        fg=typer.colors.YELLOW
+    )
+    rootdir = typer.prompt(
+        "\nEnter the directory path ('rootdir') where the config and output files will be saved",
+        default='.',
+        type=Path
+    )
+    datafile = typer.prompt(
+        "\nAnd the file path to the dataset (path/to/data.csv)",
+        type=Path
+    )
+
+    class Problem(str, Enum):
+        classification = 'c'
+        regression = 'r'
+    problem = typer.prompt(
+        "\nWhat is the problem type? It should be either classification (c) or regression (r)",
+        type=Problem
+    )
+    problem = problem.name
+
+    target_col = typer.prompt(
+        "\nThe name of the column containing target values or leave the default as the last column of the dataset",
+        type=str,
+        default=''
+    )
+    target_col = None if target_col == '' else target_col
+
+    seed = typer.prompt(
+        "\nParameter 'seed' is used to control reproducibiity. "
+        "For reproducible results, set it to any integer value",
+        type=int,
+        default=-1
+    )
+    seed = None if seed == -1 else seed
+
+    general_conf = GeneralConfig(rootdir, datafile, None, problem, target_col, seed)
+
+    typer.secho("\n# Feature selection options #", fg=typer.colors.GREEN)
+    enable_fs = typer.confirm(
+        "\nThe purpose of this step is to select only the most relevant meta-features when building the meta-dataset. "
+        "Do you want to enable it (recommended yes)",
+        default=True,
+        prompt_suffix='? '
+    )
+    if enable_fs:
+        max_n_features = typer.prompt(
+            "\nMaximum number of features to select",
+            type=int,
+            default=10
+        )
+
+        class Method(str, Enum):
+            nca = 'NCA'
+            it = 'IT'
+        method = typer.prompt(
+            "\nAvailable methods are Neighbourhood Component based ('NCA', recommended), "
+            "or Information Theoretic based ('IT'). Which one to use",
+            type=Method,
+            default=Method('NCA'),
+            prompt_suffix="? "
+        )
+        method = method.name
+
+        var_filter = typer.confirm(
+            "\nNaive variance filter removes features whose variance is smaller than a given threshold. "
+            "Enable it (recommended yes)",
+            default=True,
+            prompt_suffix='? '
+        )
+        if var_filter:
+            var_threshold = typer.prompt(
+                "\nVariance threshold (default 0, recommended)",
+                type=float,
+                default=0
+            )
+        else:
+            var_threshold = 0
+
+        corr_filter = typer.confirm(
+            "\nCorrelation filter greedly removes redundant features based on their correlation coefficient. "
+            "Enable it (recommended yes)",
+            default=True,
+            prompt_suffix='? '
+        )
+        if corr_filter:
+            corr_threshold = typer.prompt(
+                "\nCorrelation threshold (default 0.95, recommended)",
+                type=float,
+                default=0.95
+            )
+        else:
+            corr_threshold = 0.95
+
+        fs_conf = FSConfig(
+            enable_fs,
+            max_n_features,
+            method,
+            dict(),
+            var_filter,
+            var_threshold,
+            corr_filter,
+            corr_threshold
+        )
+    else:
+        fs_conf = FSConfig(enabled=False)
+
+    typer.secho("\n# Hyper-parameter optimization options #", fg=typer.colors.GREEN)
+    enable_hpo = typer.confirm(
+        "\nDo you want to enable hyper-parameter optimization",
+        default=False,
+        prompt_suffix='? '
+    )
+
+    if enable_hpo:
+        evals = typer.prompt(
+            "\nMaximum number of evaluations (per algorithm per fold)",
+            type=int,
+            default=10
+        )
+
+        timeout = typer.prompt(
+            "\nTimeout for each fold (in seconds)",
+            type=int,
+            default=90
+        )
+
+        hpo_conf = HPOConfig(enable_hpo, evals, timeout)
+    else:
+        hpo_conf = HPOConfig(enabled=False)
+
+    typer.secho("\n# ISA options #", fg=typer.colors.GREEN)
+    performance_threshold = typer.prompt(
+        "\nPerformance threshold below which an instance is considered good",
+        default='auto'
+    )
+
+    adjust_rotation = typer.confirm(
+        "\nAdjust IS orientation, such that hard instances are placed in the upper left corner",
+        prompt_suffix="? "
+    )
+
+    typer.echo("\nInstance hardness footprint parameters")
+    ih_threshold = typer.prompt(
+        "Instance easiness threshold (we recommend not to change it)",
+        default=0.4,
+        type=float
+    )
+    ih_purity = typer.prompt(
+        "Purity (we recommend not to change it)",
+        default=0.4,
+        type=float
+    )
+
+    typer.secho("\n# Hardness measures #", fg=typer.colors.GREEN)
+
+    default_measures = {
+        'classification': [
+            'kDN',
+            'DCP',
+            'TD_P',
+            'TD_U',
+            'CL',
+            'CLD',
+            'N1',
+            'N2',
+            'LSC',
+            'LSR',
+            'Harmfulness',
+            'Usefulness',
+            'F1'
+        ],
+        'regression': [
+            'C4',
+            'L1',
+            'S1',
+            'S2',
+            'S3',
+            'FO',
+            'POT',
+            'HB',
+            'EDS',
+            'EZR'
+        ]
+    }
+
+    typer.echo(f"\nSuggested measures for {problem} are:")
+
+    isa_conf = ISAConfig(
+        'python',
+        performance_threshold,
+        adjust_rotation,
+        ih_threshold,
+        ih_purity
+    )
+
+    typer.secho(general_conf, fg=typer.colors.RED)
+    typer.secho(fs_conf, fg=typer.colors.RED)
+    typer.secho(hpo_conf, fg=typer.colors.RED)
+    typer.secho(isa_conf, fg=typer.colors.RED)
+
 
 @cli.command("demo")
 def launch_demo():
@@ -141,8 +363,7 @@ def launch_app(
     if problem == 'classification':
         app_viz = ClassificationApp(workspace)
     elif problem == 'regression':
-        # app_viz = AppRegression(workspace)
-        raise NotImplementedError("Regression problems not yet supported. Coming soon!")
+        app_viz = RegressionApp(workspace)
     else:
         _logger.error(f"Unknown problem type '{problem}'.")
         raise typer.Exit(1)
@@ -245,8 +466,9 @@ def run(
 
     if isa:
         if conf.fs.enabled:
+            min_n_features = 3
             n_feat_cols = len(df_metadata.filter(regex='^feature_').columns)
-            if n_feat_cols > conf.fs.max_n_features:
+            if n_feat_cols > min_n_features:
                 _logger.info("Feature selection on")
 
                 df_metadata.to_csv(rootdir_path / 'metadata_full.csv')
@@ -282,10 +504,12 @@ def run(
                     **fs_params
                 )
 
-                if len(selected) < 3:
-                    _logger.warning(f"An insufficient number of features has been selected ({len(selected) < 3}). "
-                                    f"An attempt with a pre-fixed number (10) of features to be selected will be made")
-                    fs_params = {**fs_params, **{'n_features_to_select': 10}}
+                if len(selected) < min_n_features:
+                    _logger.warning(
+                        f"An insufficient number of features has been selected ({len(selected) < min_n_features}). "
+                        f"An attempt with the minimum number of features to be selected ({min_n_features}) will be made"
+                    )
+                    fs_params = {**fs_params, **{'n_features_to_select': min_n_features}}
                     selected = filtfeat(
                         X=X[:, idx],
                         Y=Y,
@@ -300,10 +524,11 @@ def run(
 
                 df_metadata = pd.concat([df_features[selected], df_algo], axis=1)
                 df_metadata.to_csv(rootdir_path / 'metadata.csv')
+            elif n_feat_cols == min_n_features:
+                _logger.info(f"Skipping feature selection: cannot select less than {min_n_features} features.")
             else:
-                _logger.info("Skipping feature selection: "
-                             "number of features already satisfied "
-                             f"({n_feat_cols} <= max_n_features ({conf.fs.max_n_features}))")
+                _logger.error(f"Metadata does not contain the minimum number of features ({min_n_features})!")
+                raise typer.Exit(1)
         else:
             _logger.info("Feature selection off")
 
@@ -312,10 +537,16 @@ def run(
         if isa_engine == 'python':
             # changes ISA 'perf':'epsilon' option
             epsilon = conf.isa.performance_threshold
+            abs_perf = True
             if epsilon == 'auto':
-                n_classes = df_dataset.iloc[:, -1].nunique()
-                epsilon = loss_threshold(n_classes, metric=conf.algos.metric)
-            other = {'perf': {'epsilon': epsilon}}
+                if problem == 'classification':
+                    n_classes = df_dataset.iloc[:, -1].nunique()
+                    epsilon = loss_threshold(n_classes, metric=conf.algos.metric)
+                elif problem == 'regression':
+                    epsilon = np.median(np.quantile(df_metadata.filter(regex="^algo_"), q=0.9, axis=0))
+                    # epsilon = 0.3
+                    # abs_perf = False
+            other = {'perf': {'epsilon': epsilon, 'AbsPerf': abs_perf}}
 
             _logger.info("Adding gaussian noise to the metadata for numerical stability.")
             df_features = df_metadata.filter(regex='^feature_')
@@ -345,7 +576,7 @@ def run(
                 best=[_empty_footprint()],
                 algolabels=['instance_easiness']
             )
-            model.trace.summary = model.trace.summary.append(ih_summary)
+            model.trace.summary = pd.concat([model.trace.summary, ih_summary])
 
             # Save footprints and models
             save_footprint(ih_fp, rootdir_path, 'instance_easiness')
@@ -359,7 +590,7 @@ def run(
             integrator.run_matilda_module(rootdir=rootdir_path)
         else:
             _logger.error(f"Unknown ISA engine '{repr(isa_engine)}'.")
-            typer.Exit(1)
+            raise typer.Exit(1)
 
     end = time.time()
     elapsed_time = int(end - start)
