@@ -22,7 +22,8 @@ from pyispace.utils import scriptcsv
 
 from .classification import ClassifiersPool
 from .feature_selection import featfilt
-from .measures import ClassificationMeasures
+from .measures import ClassificationMeasures, RegressionMeasures
+from .regression import RegressorsPool
 from .structures import Configurations
 
 metadata_file = 'metadata.csv'
@@ -64,9 +65,8 @@ def build_metadata(
         measures = ClassificationMeasures(data, target_col=target_col)
         learners = ClassifiersPool(data, labels_col=target_col)
     elif problem == 'regression':
-        # measures = ...
-        # learners = ...
-        raise NotImplementedError("Regression problems not yet supported. Coming soon!")
+        measures = RegressionMeasures(data, target_col=target_col)
+        learners = RegressorsPool(data, output_col=target_col)
     else:
         raise ValueError(f"Unknown type of problem: '{problem}'.")
 
