@@ -8,12 +8,15 @@ import logging
 import numpy as np
 import pandas as pd
 from scipy.sparse.csgraph import minimum_spanning_tree
+from scipy.stats import iqr
 from sklearn import tree
 from sklearn.calibration import CalibratedClassifierCV
+from sklearn.linear_model import LinearRegression
 from sklearn.metrics import DistanceMetric
 from sklearn.model_selection import GridSearchCV
 from sklearn.naive_bayes import GaussianNB
 from sklearn.neighbors import NearestNeighbors
+from sklearn.preprocessing import StandardScaler
 
 from . import get_seed
 from .base import Measures
@@ -198,7 +201,7 @@ class ClassificationMeasures(Measures):
             parameters = {'ccp_alpha': np.linspace(0.001, 0.1, num=100)}
             dtc = tree.DecisionTreeClassifier(criterion='gini', random_state=seed)
             clf = GridSearchCV(dtc, parameters, n_jobs=-1)
-            clf = clf.fit(self.X.values, self.y.values)
+            clf.fit(self.X.values, self.y.values)
             ccp_alpha = clf.best_params_['ccp_alpha']
 
         self.dtc_pruned = tree.DecisionTreeClassifier(criterion='gini', ccp_alpha=ccp_alpha, random_state=seed)
@@ -222,7 +225,7 @@ class ClassificationMeasures(Measures):
 
         self.dot = None
 
-    def k_disagreeing_neighbors(self, k=10, distance='gower') -> np.ndarray:
+    def k_disagreeing_neighbors(self, k: int = 10, distance: str = 'gower') -> np.ndarray:
         r"""
         k-Disagreeing Neighbors (kDN) gives the percentage of the :math:`k` nearest neighbors of :math:`\\mathbf x_i`
         which do not share its label.
@@ -255,7 +258,7 @@ class ClassificationMeasures(Measures):
     def disjunct_size(self) -> np.ndarray:
         data = self.data.copy()
 
-        data['leaf_id'] = self.dtc.apply(self.X)
+        data['leaf_id'] = self.dtc.apply(self.X.values)
         df_count = data.groupby('leaf_id').count().iloc[:, 0].to_frame('count').subtract(1)
         data = data.join(df_count, on='leaf_id')
         DS = data['count'].divide(data['count'].max())
@@ -377,6 +380,18 @@ class ClassificationMeasures(Measures):
         return (1 - CLD) / 2
 
     def minority_value(self) -> np.ndarray:
+        r"""
+        Also measures the skewness of :math:`\mathbf x_i`'s class, with an alternative formulation based on the
+        imbalance ratio of the class the instance belongs to:
+
+        .. math::
+
+            MV(\mathbf x_i) = 1 - \frac{\sharp\{\mathbf x_j | y_j = y_i\}}{\max_{y_j \in \mathcal Y}
+            \left\{\sharp\{\mathbf x_j|\mathbf x_j \in y_j\}\right\}}
+
+        Returns:
+            array-like: :math:`MV(\mathbf x_i)`
+        """
         mv_class = self.data.groupby(self.target_col).count().iloc[:, 0]
         mv_class = mv_class.divide(mv_class.max())
 
@@ -384,11 +399,26 @@ class ClassificationMeasures(Measures):
         return labels.apply(lambda c: 1 - mv_class[c]).values
 
     def class_balance(self) -> np.ndarray:
+        r"""
+        Measures the skewness of :math:`\mathbf x_i`'s class as:
+
+        .. math::
+
+            CB(\mathbf x_i) = \left( 1 - \frac{ \sharp\{\mathbf x_j | y_j = y_i\}}{n} + \frac{1}{C} \right)
+            \left( \frac{C}{C+1}\right)
+
+        where :math:`n` is the number of instances in the dataset :math:`\mathcal D` and :math:`C` is the number of
+        classes the problem has.
+
+        Returns:
+            array-like: :math:`CB(\mathbf x_i)`
+        """
         cb_class = self.data.groupby(self.target_col).count().iloc[:, 0]
         cb_class = cb_class.divide(self.N) - 1 / len(self.y.unique())
 
         labels = self.y
-        return (1 - labels.apply(lambda c: cb_class[c]).values) / 1.5
+        n_classes = self.y.nunique()
+        return (1 - labels.apply(lambda c: cb_class[c]).values) * n_classes / (n_classes + 1)
 
     def borderline_points(self) -> np.ndarray:
         r"""
@@ -556,7 +586,8 @@ class ClassificationMeasures(Measures):
 
         .. math::
 
-            H(\mathbf x_i) = \frac{\sharp\{\mathbf x_j | ne(\mathbf x_j) = \mathbf x_i \}}{|D|-1}
+            H(\mathbf x_i) = \frac{\sharp\{\mathbf x_j | ne(\mathbf x_j) = \mathbf x_i \}}{\sharp\{\mathbf x_j|y_i
+            \neq y_j\}}
 
         If :math:`\mathbf x_i` is the nearest enemy of many instances, this indicates it is harder to classify and its
         harmfulness will be higher.
@@ -582,7 +613,7 @@ class ClassificationMeasures(Measures):
         n_class = self.y.value_counts()
 
         H = np.sum(np.reshape(np.arange(self.N), (self.N, -1)) == ne, axis=1)
-        return H / n_class[self.y].values
+        return H / (self.N - n_class[self.y].values)
 
     def usefulness(self, distance='gower') -> np.ndarray:
         r"""
@@ -591,7 +622,7 @@ class ClassificationMeasures(Measures):
         .. math::
 
             U(\mathbf x_i) = 1 - \frac{\sharp \{\mathbf x_j| d(\mathbf x_i,\mathbf x_j) < d(\mathbf x_j,
-            ne(\mathbf x_j))\}}{|D|-1}
+            ne(\mathbf x_j))\}}{{\sharp\{\mathbf x_j|y_i=y_j\}}}
 
         If :math:`\mathbf x_i` is easy to classify, it will be close to many examples from its class and therefore will
         be more useful. We take the complement of this measure as output.
@@ -665,7 +696,7 @@ class ClassificationMeasures(Measures):
             for f in features:
                 m1 = maxmin(sub_df[f].values, sub_df[self.target_col].values)
                 m2 = minmax(sub_df[f].values, sub_df[self.target_col].values)
-                indicator += sub_df[f].between(m1, m2, inclusive=False) * 1
+                indicator += sub_df[f].between(m1, m2, inclusive='neither') * 1
             F1 = F1.add(indicator.divide(n_features), fill_value=0)
 
         F1 = F1.divide(len(classes) - 1)
@@ -704,3 +735,332 @@ class ClassificationMeasures(Measures):
         if self.dot is None:
             self.dot = self.do_t()
         return self.dot.max(axis=1).values
+
+
+class RegressionMeasures(Measures):
+    """
+    Hardness measures for regression. It provides separate methods to compute each measure.
+
+    Args:
+        data (pd.DataFrame): a dataframe where each line is an instace and columns are features. One column should
+            contain the target value. The name of the column with targets can be set with parameter `labels_col`
+        target_col (str): name of the column that contains the labels of the instances (default None - uses the
+            last column)
+    """
+
+    # measure_name : function_name
+    _measures_dict = {
+        'C4': 'collective_feature_efficiency',
+        'L1': 'linear_absolute_error',
+        'S1': 'output_distribution',
+        'S2': 'input_distribution',
+        'S3': 'error_nn_regressor',
+        'FO': 'feature_outlier',
+        'POT': 'target_outlier',
+        'HB': 'histogram_bin',
+        'EDS': 'stump_error',
+        'EZR': 'zero_rule_error',
+        'DS': 'disjunct_size',
+        'TD_P': 'tree_depth_pruned',
+        'TD_U': 'tree_depth_unpruned',
+        'D': 'density',
+        'CC': 'clustering_coefficient'
+    }
+
+    logger = logging.getLogger(__name__)
+
+    def __init__(self, data: pd.DataFrame, target_col=None, ccp_alpha=None, eps=0.15):
+        if target_col is None:
+            self.target_col = data.columns[-1]
+            self.y = data.iloc[:, -1].values
+        else:
+            self.target_col = target_col
+            self.y = data[target_col].values
+
+        self.data = data.reset_index(drop=True)
+        self.X = data.drop(columns=self.target_col).values
+        self.N, self.n_feat = self.X.shape
+
+        # scaled data
+        self.X_scaled = StandardScaler().fit_transform(self.X)
+        self.y_scaled = (self.y - self.y.mean()) / self.y.std()
+
+        self.dist_matrix_gower = gower_distance(data.drop(columns=self.target_col))
+        # delta = np.diag(-np.ones(self.dist_matrix_gower.shape[0]))
+        # self.indices_gower = np.argsort(self.dist_matrix_gower + delta, axis=1)
+        # self.distances_gower = np.sort(self.dist_matrix_gower, axis=1)
+
+        # use it for random_state
+        seed = get_seed()
+
+        # Decision Tree Regressor
+        dtr = tree.DecisionTreeRegressor(min_samples_split=2, criterion='squared_error', random_state=seed)
+        self.dtr = dtr.fit(self.X_scaled, self.y_scaled)
+
+        # Decision Tree Regressor Pruned
+        if ccp_alpha is None:
+            parameters = {'ccp_alpha': np.linspace(0.001, 0.1, num=100)}
+            dtr = tree.DecisionTreeRegressor(criterion='squared_error', random_state=seed)
+            clf = GridSearchCV(dtr, parameters, n_jobs=-1)
+            clf.fit(self.X_scaled, self.y_scaled)
+            ccp_alpha = clf.best_params_['ccp_alpha']
+
+        self.dtr_pruned = tree.DecisionTreeRegressor(criterion='squared_error', ccp_alpha=ccp_alpha, random_state=seed)
+        self.dtr_pruned = self.dtr_pruned.fit(self.X_scaled, self.y_scaled)
+
+        # Build adjacency matrix to generate graph dataset
+        self.adjacency_matrix = np.where(self.dist_matrix_gower >= eps, 1, 0)
+        np.fill_diagonal(self.adjacency_matrix, 0)
+
+    def collective_feature_efficiency(self, epsilon: float = 0.1) -> np.ndarray:
+        r"""
+        This measure starts by identifying the feature with the highest correlation to the output in :math:`D`. All
+        examples with a small residual value (:math:`|\varepsilon_i| \leq 0.1`) after a linear fit between this feature
+        and the target attribute are removed. Then, the most correlated feature to the remaining data points is found
+        and the previous process is repeated until all features have been analyzed or no example remains. For an
+        instance :math:`\mathbf x_i`, we take the round :math:`l_i` where it is removed from the analysis, or output
+        the value 1 if it remains.
+
+        .. math::
+
+            C_4(\mathbf x_i) = \frac{l_i}{m+1}
+
+        if :math:`\mathbf x_i` is removed in one of the :math:`C_4` rounds. Otherwise, the :math:`l_i` value is set as
+        :math:`m+1`. Higher :math:`C_4` values are obtained for harder instances, which require more features to get a
+        linear fit. The :math:`C_4` values range in :math:`\left[\frac{1}{m+1},1\right]`.
+
+        Args:
+            epsilon (float): residual value threshold
+
+        Returns:
+            array-like: Collective feature efficiency values per instance.
+        """
+        C4 = np.ones(self.y.shape)
+
+        for i in range(self.n_feat):
+            x_feat = self.X_scaled[:, i].reshape(-1, 1)
+            lr = LinearRegression(fit_intercept=False)
+            lr.fit(x_feat, self.y_scaled)
+            y_pred = lr.predict(x_feat)
+            resid = np.abs(self.y_scaled - y_pred)
+            mask = resid <= epsilon
+            C4[mask] = (i + 1) / (self.n_feat + 1)
+
+        return C4
+
+    def linear_absolute_error(self) -> np.ndarray:
+        r"""
+        First a statistical model of a Multiple Linear Regression is fit to :math:`D`. For each :math:`\mathbf x_i`
+        a residual or error :math:`\varepsilon_i` in relation to the actual output :math:`y_i` can be measure and
+        :math:`L_1` is given as:
+
+        .. math::
+
+            L_1(\mathbf x_i) = |\varepsilon_i|.
+
+        Larger values are attained for harder instances.
+
+        Returns:
+            array-like: Absolute error after linear fit values per instance.
+        """
+        lr = LinearRegression(fit_intercept=False)
+        lr.fit(self.X_scaled, self.y_scaled)
+        y_pred = lr.predict(self.X_scaled)
+        L1 = np.abs(self.y_scaled - y_pred)
+        L1_normalized = 1 - 1/(1 + L1)
+        return L1_normalized
+
+    def output_distribution(self) -> np.ndarray:
+        r"""
+        As in :math:`N_1` for classification, first a MST is generated from input data. Next :math:`S_1` monitors
+        whether the instances joined in the MST have similar output values. An instance :math:`\mathbf x_i` can have
+        multiple neighbors in the MST. So, take the average of the differences between their outputs:
+
+        .. math::
+
+            S_1(\mathbf x_i) = \frac{1}{n_i}\sum_{i:j \in MST} |y_i - y_j|
+
+        where :math:`n_i` is the number of neighbors of :math:`\mathbf x_i` in the MST. Higher values will be obtained
+        for harder instances, which are connected to examples with dissimilar outputs.
+
+        Returns:
+            array-like: Output distribution measure values per instance.
+        """
+        Tcsr = minimum_spanning_tree(self.dist_matrix_gower)  # noqa
+
+        adjacency = Tcsr.maximum(Tcsr.T)
+        S1 = adjacency.sum(axis=1) / np.sum(adjacency > 0, axis=1)
+        S1 = np.squeeze(np.asarray(S1))
+
+        return S1
+
+    def input_distribution(self) -> np.ndarray:
+        r"""
+        :math:`S_2` first orders the data points according to their output values :math:`y_i` and then computes the
+        Euclidean distance between pairs of examples that are neighbors. In the ordering, each element will have one or
+        two neighbor examples. For two neighbors, the average should be taken. Otherwise, if the example is the first or
+        the last in the ordering, the unique neighborhood difference is output. Given that
+        :math:`y_1 \leq y_2 \leq \ldots \leq y_n`, that is, that the examples are already ordered according to their
+        output values, we have:
+
+        .. math::
+
+            S_2(\mathbf x_i) = \frac{d(\mathbf x_i,\mathbf x_{i-1})+d(\mathbf x_i,\mathbf x_{i+1})}{2}
+
+        for :math:`2<i<n-1`. For :math:`i=1`:
+
+        .. math::
+
+            S_2(\mathbf x_1) = d(\mathbf x_1,\mathbf x_2)
+
+        And for :math:`i=n`:
+
+        .. math::
+
+            S_2(\mathbf x_n) = d(\mathbf x_n,\mathbf x_{n-1})
+
+        Larger values are obtained for harder instances.
+
+        Returns:
+            array-like: Input distribution measure values per instance.
+        """
+        idx = np.argsort(self.y)
+        y_ord = self.y_scaled[idx]
+        X_ord = self.X_scaled[idx, :]
+
+        d1 = np.linalg.norm(np.diff(X_ord, axis=0), ord=2, axis=1)
+        d2 = np.linalg.norm(np.diff(X_ord[::-1, :], axis=0)[::-1, :], ord=2, axis=1)
+
+        S2 = np.zeros(y_ord.shape[0])
+        aux = np.zeros(y_ord.shape[0])
+        aux[1:] += d1
+        aux[:-1] += d2
+        aux[1:-1] *= 1 / 2
+        S2[idx] = aux
+
+        return 1 - 1 / (1 + S2)
+
+    def error_nn_regressor(self) -> np.ndarray:
+        r"""
+        Calculates the squared error (SE) of a *1-nearest neighbor regressor* (NN), using *leave-one-out*. The 1-NN
+        regressor looks for the training example :math:`\mathbf x_i` most similar to the new example in the input space
+        and assigns to it the same output :math:`y_i`.
+
+        .. math::
+
+            S_3(\mathbf x_i) = (NN(\mathbf x_i)-y_i)^2
+
+        where :math:`NN(\mathbf x_i)` represents the 1-nearest neighbor prediction for :math:`\mathbf x_i`. Larger
+        values are observed for harder instances.
+
+        Returns:
+            array-like: Squared error of 1-nearest neighbor values per instance.
+        """
+        nn = NearestNeighbors(n_neighbors=1, metric='precomputed').fit(self.dist_matrix_gower)
+        nn_arr = nn.kneighbors_graph(mode='distance').toarray()
+        _, neighs = np.nonzero(nn_arr)
+        S3 = (self.y_scaled[neighs] - self.y_scaled) ** 2
+
+        return 1 - np.exp(-S3 / S3.std())
+
+    def feature_outlier(self) -> np.ndarray:
+        IQR = iqr(self.X_scaled, axis=0)
+        Q1 = np.quantile(self.X_scaled, 0.25, axis=0)
+        Q3 = np.quantile(self.X_scaled, 0.75, axis=0)
+        lbound = Q1 - 1.5 * IQR
+        ubound = Q3 + 1.5 * IQR
+
+        FO = np.mean((self.X_scaled < lbound) | (self.X_scaled > ubound), axis=1)
+        return FO
+
+    def target_outlier(self) -> np.ndarray:
+        IQR_t = iqr(self.y)
+        Q1_t = np.quantile(self.y, 0.25)
+        Q3_t = np.quantile(self.y, 0.75)
+        lbound_t = Q1_t - 1.5 * IQR_t
+        ubound_t = Q3_t + 1.5 * IQR_t
+
+        POT = (self.y < lbound_t) | (self.y > ubound_t)
+        POT = POT * 1
+        return POT
+
+    def histogram_bin(self) -> np.ndarray:
+        r"""
+        First a histogram of the normalized values of :math:`\mathbf y` is taken. Next, we output the complement of the
+        percentage of instances that have their labels placed in the same bin as :math:`\mathbf x_i`. Easy instances
+        will tend to share labels with many other instances in :math:`D` than hard instances.
+
+        Returns:
+            array-like: Histogram bin measure values per instance.
+        """
+        n_bin, bins = np.histogram(self.y_scaled, density=False, bins='auto')
+        bin_idx = np.digitize(self.y_scaled, bins=bins, right=False)
+        HB = 1 - n_bin[np.clip(bin_idx, a_max=len(n_bin), a_min=None) - 1] / n_bin.sum()
+        return HB
+
+    def stump_error(self) -> np.ndarray:
+        stump = tree.DecisionTreeRegressor(max_depth=1, criterion='squared_error').fit(self.X, self.y)
+        EDS = (self.y - stump.predict(self.X)) ** 2
+        return 1 - np.exp(-EDS / EDS.std())
+
+    def zero_rule_error(self) -> np.ndarray:
+        zero_r = (self.y - np.median(self.y)) ** 2
+        return 1 - np.exp(-zero_r / zero_r.std())
+
+    def disjunct_size(self) -> np.ndarray:
+        data = self.data.copy()
+
+        data['leaf_id'] = self.dtr.apply(self.X)
+        df_count = data.groupby('leaf_id').count().iloc[:, 0].to_frame('count').subtract(1)
+        data = data.join(df_count, on='leaf_id')
+        DS = data['count'].divide(data['count'].max())
+
+        return 1 - DS.values
+
+    def tree_depth_unpruned(self) -> np.ndarray:
+        r"""
+        Tree Depth (TD) returns the depth of the leaf node that predicts :math:`\mathbf x_i` in a  decision tree,
+        normalized by the maximum depth of the tree built from :math:`D`:
+
+        .. math::
+
+            TD(\mathbf x_i) = \frac{depth(\mathbf x_i)}{\max(depth(D))}
+
+        There are two versions of this measure, using pruned (:math:`TD_P(\mathbf x_i)`)
+        and unpruned (:math:`TD_U(\mathbf x_i)`) decision trees. Instances harder to predict tend to be placed
+        at deeper levels of the trees and present higher :math:`TD` values.
+
+        Returns:
+            array-like: :math:`TD_U(\mathbf x_i)`
+        """
+        return pd.DataFrame(self.X_scaled).apply(lambda x: self.dtr.decision_path([x]).sum() - 1,
+                                                 axis=1, raw=True).values / self.dtr.get_depth()
+
+    def tree_depth_pruned(self) -> np.ndarray:
+        r"""
+        Tree Depth (TD) returns the depth of the leaf node that predicts :math:`\mathbf x_i` in a  decision tree,
+        normalized by the maximum depth of the tree built from :math:`D`:
+
+        .. math::
+
+            TD(\mathbf x_i) = \frac{depth(\mathbf x_i)}{\max(depth(D))}
+
+        There are two versions of this measure, using pruned (:math:`TD_P(\mathbf x_i)`)
+        and unpruned (:math:`TD_U(\mathbf x_i)`) decision trees. Instances harder to predict tend to be placed
+        at deeper levels of the trees and present higher :math:`TD` values.
+
+        Returns:
+            array-like: :math:`TD_P(\mathbf x_i)`
+        """
+        return pd.DataFrame(self.X_scaled).apply(lambda x: self.dtr_pruned.decision_path([x]).sum() - 1,
+                                                 axis=1, raw=True).values / self.dtr_pruned.get_depth()
+
+    def density(self) -> np.ndarray:
+        nodes_degrees = np.sum(self.adjacency_matrix, axis=1)
+        return nodes_degrees / (self.N - 1)
+
+    def clustering_coefficient(self) -> np.ndarray:
+        degrees = np.sum(self.adjacency_matrix, axis=1)
+        unormalized = np.divide(2 * np.sum(self.adjacency_matrix * (degrees - 1), axis=1),
+                                degrees * (degrees - 1))
+        return 1 - np.divide(unormalized, self.N)
