@@ -53,18 +53,46 @@ def absolute_error(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
 
 def squared_error(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
     """
-        Calculates the squared error value for each instance.
+    Calculates the squared error value for each instance.
 
-        Args:
-            y_true (array-like): true value
-            y_pred (array-like): predicted value
+    Args:
+        y_true (array-like): true value
+        y_pred (array-like): predicted value
 
-        Returns:
-            array-like: squared error values
-        """
+    Returns:
+        array-like: squared error values
+    """
     return (y_pred - y_true) ** 2
 
 
+def normalized_absolute_error(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
+    r"""
+    Calculates the normalized absolute error value for each instance.
+
+    Args:
+        y_true (array-like): true value
+        y_pred (array-like): predicted value
+
+    Returns:
+        array-like: normalized absolute error values
+    """
+    return np.abs(y_true - y_pred) / np.abs(y_true).mean()
+
+
+def normalized_squared_error(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
+    """
+    Calculates the normalized squared error value for each instance.
+
+    Args:
+        y_true (array-like): true value
+        y_pred (array-like): predicted value
+
+    Returns:
+        array-like: normalized squared error values
+    """
+    return (y_true - y_pred)**2 * len(y_true) / y_true.dot(y_true.T)
+
+
 def loss_threshold(n_classes: int, metric: str = 'logloss', eps: float = 1e-3):
     """
     Calculates the maximum threshold below which the metric indicates a correct classification of the instance. It is
