@@ -323,6 +323,265 @@ class AppClassification(BaseApp):
                                                        sizing_mode='stretch_height'))
 
 
+class AppRegression(BaseApp):
+    def __init__(self, workspace: Workspace):
+        super().__init__(workspace)
+        hv.extension('bokeh')
+        hv.renderer('bokeh').theme = light_theme
+
+        df_dataset = workspace.data.copy()
+        df_metadata = workspace.extended_metadata
+        df_is = workspace.is_coordinates
+        df_footprint = workspace.footprints
+
+        if len(df_dataset.columns) > 3:
+            X = df_dataset.iloc[:, :-1]
+            y = df_dataset.iloc[:, -1]
+
+            pca = PCA(n_components=2)
+            X_embedded = pca.fit_transform(X)
+
+            df = pd.DataFrame(X_embedded, columns=['Component1', 'Component2'], index=X.index)
+            df_dataset = pd.concat([df, y], axis=1)
+
+        data = df_is.join(df_dataset)
+        self.data = data.join(df_metadata)
+
+        self.data_dims = df_dataset.columns.to_list()
+        self.target = self.data_dims[2]
+        self.meta_dims = df_metadata.columns.to_list()
+
+        # Scatter kdims and vdims
+        is_cols = df_is.columns.to_list()
+        self.is_kdims = is_cols[0]
+        self.is_vdims = [is_cols[1], self.target] + self.meta_dims
+        self.data_kdims = self.data_dims[0]
+        self.data_vdims = [self.data_dims[1], self.target] + self.meta_dims
+
+        # Panel widgets
+        self.w_color = pn.widgets.Select(options=self.meta_dims + [self.target], value=self.meta_dims[0])
+        self.w_color_range = pn.widgets.RangeSlider(start=0, end=20, value=(0, 5), step=0.5)
+        self.w_checkbox = pn.widgets.Checkbox(name='manual colorbar range', value=False)
+        self.w_footprint_on = pn.widgets.Checkbox(name='draw footprint area', value=True)
+
+        # Data explorer tab widgets
+        self.w_var_y = pn.widgets.Select(name='y-axis', options=self.all_cols, value='z_2')
+        self.w_var_x = pn.widgets.Select(name='x-axis', options=self.all_cols, value='z_1')
+        self.w_var_c = pn.widgets.Select(name='Color', options=self.all_cols, value=self.output_col)
+        self.w_var_cm = pn.widgets.RadioButtonGroup(name='Color map', options=['Continuous', 'Discrete'],
+                                                    button_type='default', value='Continuous')
+        self.w_var_s = pn.widgets.DiscreteSlider(name='Size', options=list(range(1, 16)), value=7)
+
+        val = 'instance_hardness' if 'instance_hardness' in df_footprint.index else ''
+        self.w_footprint_algo = pn.widgets.Select(options=df_footprint.index.unique(level='algo').to_list(), value=val)
+
+        # self.populate_tabs()
+
+    def data_space(self, c, lim, autorange_on):
+        if not autorange_on:
+            lim = (np.nan, np.nan)
+        cmap = self.cmap
+        hover_list = [(self.output_col, self.output_col),
+                      ('Component 1', 'Component1'),
+                      ('Component 2', 'Component2'),
+                      ('instance hardness', 'instance_hardness'),
+                      (' '.join(str.split(c, '_', 1)), c)]
+        hover_list = list(dict.fromkeys(hover_list))
+        tooltips = [("index", "$index")] + [(s[0], '@' + s[1]) for s in hover_list]
+        hover = HoverTool(tooltips=tooltips)
+        scatter1 = hv.Scatter(self.data,
+                              kdims=self.data_kdims,
+                              vdims=self.data_vdims
+                              ).opts(responsive=True,
+                                     aspect=1.1,
+                                     color=c,
+                                     cmap=cmap,
+                                     show_grid=True,
+                                     # marker=self.marker,
+                                     tools=['lasso_select', 'box_select', hover],
+                                     size=7,
+                                     framewise=True,
+                                     colorbar=True,
+                                     clim=lim)
+        return scatter1
+
+    def instance_space(self, c, lim, autorange_on):
+        if not autorange_on:
+            lim = (np.nan, np.nan)
+        cmap = self.cmap
+        hover_list = [(self.output_col, self.output_col),
+                      ('z1', 'z_1'),
+                      ('z2', 'z_2'),
+                      ('instance hardness', 'instance_hardness'),
+                      (' '.join(str.split(c, '_', 1)), c)]
+        hover_list = list(dict.fromkeys(hover_list))
+        tooltips = [("index", "$index")] + [(s[0], '@' + s[1]) for s in hover_list]
+        hover = HoverTool(tooltips=tooltips)
+        scatter2 = hv.Scatter(self.data,
+                              kdims=self.is_kdims,
+                              vdims=self.is_vdims
+                              ).opts(responsive=True,
+                                     aspect=1.1,
+                                     color=c,
+                                     cmap=cmap,
+                                     show_grid=True,
+                                     # marker=self.marker,
+                                     tools=['lasso_select', 'box_select', hover],
+                                     size=7,
+                                     framewise=True,
+                                     colorbar=True,
+                                     clim=lim)
+        return scatter2
+
+    def populate_tabs(self):
+        mask = self.select_instances()
+        if not mask.any():
+            mask = ~mask
+        df_selection = self.workspace.data[mask].copy()
+
+        output_col = df_selection.columns[-1]
+        df_selection.loc[:, output_col] = df_selection.iloc[:, -1].apply(lambda x: str(x))
+        cols = df_selection.columns[:-1].to_list()
+
+        # metadata selection tab
+        df_feat = self.workspace.metadata[mask].filter(regex='^feature_').melt(var_name='feature')
+        boxplot_feat = hv.BoxWhisker(df_feat, kdims='feature', vdims='value').opts(responsive=True,
+                                                                                   aspect=3,
+                                                                                   xrotation=45,
+                                                                                   show_grid=True,
+                                                                                   box_fill_color=dim('feature').str(),
+                                                                                   cmap='glasbey_cool',
+                                                                                   tools=['hover'],
+                                                                                   ylim=(0, 1))
+
+        df_algo = self.workspace.metadata[mask].filter(regex='^algo_').melt(var_name='algorithm', value_name='logloss')
+        boxplot_algo = hv.BoxWhisker(df_algo, kdims='algorithm', vdims='logloss').opts(responsive=True,
+                                                                                       aspect=3,
+                                                                                       xrotation=45,
+                                                                                       show_grid=True,
+                                                                                       box_fill_color=dim(
+                                                                                           'algorithm').str(),
+                                                                                       cmap='glasbey_cool',
+                                                                                       tools=['hover'])
+
+        self._inner_tabs[0] = (self._tabs_inner_id[0],
+                               pn.Column('# Distribution of the meta-features', pn.Row(boxplot_feat),
+                                         '# Distribution of the algorithms', pn.Row(boxplot_algo),
+                                         sizing_mode='stretch_both'))
+
+        # integer_cols = [col for col in cols if is_integer_dtype(df_selection[col])]
+        # float_cols = [col for col in cols if is_float_dtype(df_selection[col])]
+        #
+        # w_var_box = pn.widgets.MultiSelect(name='Features', value=[], options=float_cols)
+        # w_var_hist = pn.widgets.Select(name='Feature', options=cols)
+        # w_var_bar = pn.widgets.Select(name='Feature', options=integer_cols)
+        #
+        # @pn.depends(feat_list=w_var_box.param.value)
+        # def make_box(feat_list):
+        #     return hv.BoxWhisker(df_selection[feat_list + [output_col]].melt(id_vars=output_col, var_name='feature'),
+        #                          kdims=['feature', output_col],
+        #                          vdims='value').opts(responsive=True,
+        #                                              aspect=2.5,
+        #                                              xrotation=45,
+        #                                              show_grid=True,
+        #                                              box_fill_color=dim(output_col).str(),
+        #                                              box_cmap='Set1',
+        #                                              tools=['hover'],
+        #                                              fontscale=1.2)
+        #
+        # @pn.depends(var=w_var_bar.param.value)
+        # def make_bar(var):
+        #     if var is None:
+        #         bar = hv.Bars([])
+        #     else:
+        #         kdims = [var, output_col]
+        #         s = df_selection[kdims].value_counts()
+        #         s.name = 'Count'
+        #         s = s.to_frame()
+        #         bar = hv.Bars(s)
+        #     tooltips = [(output_col, f'@{output_col}'), ('Count', '@Count')]
+        #     bar.opts(responsive=True,
+        #              aspect=2.5,
+        #              multi_level=False,
+        #              stacked=False,
+        #              show_grid=True,
+        #              fontscale=1.2,
+        #              tools=[HoverTool(tooltips=tooltips)])
+        #     return bar
+        #
+        # @pn.depends(var=w_var_hist.param.value)
+        # def make_hist(var):
+        #     classes = df_selection.iloc[:, -1].unique()
+        #     a_min = df_selection[var].min()
+        #     a_max = df_selection[var].max()
+        #     hist = [hv.Histogram(np.histogram(df_selection[df_selection[output_col] == c][var],
+        #                                       range=(a_min, a_max), bins=40), label=str(c))
+        #             for c in classes]
+        #     if len(hist) == 0:
+        #         hist = hv.Histogram([])
+        #     else:
+        #         hist = reduce(lambda h1, h2: h1 * h2, hist)
+        #     hist.opts('Histogram',
+        #               responsive=True,
+        #               aspect=2.5,
+        #               alpha=0.8,
+        #               show_grid=True,
+        #               muted_fill_alpha=0.1,
+        #               framewise=True,
+        #               tools=['hover'],
+        #               cmap='Set1',
+        #               fontscale=1.2,
+        #               xlabel=var)
+        #     return hist
+        #
+        # dmap_box = hv.DynamicMap(make_box)
+        # dmap_box.opts(framewise=True)
+        #
+        # dmap_hist = hv.DynamicMap(make_hist)
+        # dmap_hist.opts(framewise=True)
+        #
+        # dmap_bar = hv.DynamicMap(make_bar)
+        # dmap_bar.opts(framewise=True)
+        #
+        # accordion = pn.Accordion(('Boxplot', pn.Row(w_var_box, dmap_box, sizing_mode='stretch_both')),
+        #                          ('Histogram', pn.Row(w_var_hist, dmap_hist, sizing_mode='stretch_both')),
+        #                          ('Barplot', pn.Row(w_var_bar, dmap_bar, sizing_mode='stretch_both')),
+        #                          sizing_mode='stretch_both')
+        #
+        # self._inner_tabs[1] = (self._tabs_inner_id[1], accordion)
+
+        # Explorer outer tab
+        df_all = self.workspace.data[mask].copy()
+        df_all = df_all.join(self.workspace.extended_metadata, how='left')
+        df_all = df_all.join(self.workspace.is_coordinates, how='left')
+
+        @pn.depends(x=self.w_var_x.param.value, y=self.w_var_y.param.value, c=self.w_var_c.param.value,
+                    cm=self.w_var_cm.param.value, s=self.w_var_s.param.value)
+        def plot_scatter(x, y, c, cm, s, **kwargs):
+            cm = {'Discrete': 'Set1', 'Continuous': 'coolwarm'}[cm]
+            return hv.Scatter(df_all[list({x, y, c})], kdims=[x], vdims=[y, c]).opts(responsive=True,
+                                                                                     aspect=1.8,
+                                                                                     color=c,
+                                                                                     cmap=cm,
+                                                                                     show_grid=True,
+                                                                                     size=s,
+                                                                                     framewise=True,
+                                                                                     colorbar=True,
+                                                                                     tools=['hover']
+                                                                                     )
+
+        dmap_scatter = hv.DynamicMap(plot_scatter)
+        dmap_scatter.opts(framewise=True)
+
+        self._page_tabs[3] = (self._tabs_id[3], pn.Row(pn.Column(self.w_var_y,
+                                                                 self.w_var_x,
+                                                                 self.w_var_c,
+                                                                 self.w_var_cm,
+                                                                 self.w_var_s),
+                                                       dmap_scatter,
+                                                       sizing_mode='stretch_height'))
+
+
 class Demo:
     def __init__(self, datadir=None):
         hv.extension('bokeh', logo=False)
